/*
    Copyright 2007 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www->gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <pcap.h>
#include "ether_hdrlen.h"
#include "acpdump.h"

typedef uint8_t     u8;
typedef int8_t      i8;
typedef uint16_t    u16;
typedef int16_t     i16;
typedef uint32_t    u32;
typedef int32_t     i32;




#define VER                 "0.1.1"
#ifndef LITTLE_ENDIAN
    #define LITTLE_ENDIAN   0
    #define BIG_ENDIAN      1
#endif



int udp2tcp(FILE *fd, FILE *fdo, u32 caplen);
void dump(FILE *fdo, u32 sip, u16 sport, u32 dip, u16 dport, u8 *data, int datalen);
int fgetxx(FILE *fd, u32 *ret, int bits, int endian);
void std_err(void);



#pragma pack(2)
typedef struct {
    u32    sip;
    u16    sport;
    u32    dip;
    u16    dport;
    u32    seq1;
    u32    ack1;
    u32    seq2;
    u32    ack2;
} hosts_t;
#pragma pack()

hosts_t *hosts  = NULL;
int     hosts_memsize,
        hosts_len,
        if_offset;



int main(int argc, char *argv[]) {
    FILE    *fd,
            *fdo;
    u32     n;
    int     endian  = BIG_ENDIAN;
    u8      *infile,
            *outfile;

    fputs("\n"
        "Tcpdump format UDP 2 TCP converter "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stderr);

    if(argc < 3) {
        printf("\n"
            "Usage: %s <input.PCAP> <outfile.PCAP>\n"
            "\n", argv[0]);
        exit(1);
    }

    infile  = argv[1];
    outfile = argv[2];

    if(!strcmp(infile, "-")) {
        fprintf(stderr, "- open stdin\n");
        fd = stdin;
    } else {
        fprintf(stderr, "- open input file %s\n", infile);
        fd = fopen(infile, "rb");
        if(!fd) std_err();
    }

    if(!strcmp(outfile, "-")) {
        fprintf(stderr, "- open stdout\n");
        fdo = stdout;
    } else {
        fprintf(stderr, "- create output file %s\n", outfile);
        fdo = fopen(outfile, "wb");
        if(!fdo) std_err();
    }
    create_acp(fdo);

    if(fgetxx(fd, &n, 32, endian) < 0) goto quit;   // magic
    if(n == 0xa1b2c3d4) {
    } else if(n == 0xd4c3b2a1) {
        endian = LITTLE_ENDIAN;
    } else {
        fprintf(stderr, "\nError: the input file is not in tcpdump format\n");
        exit(1);
    }
    if(fgetxx(fd, &n, 16, endian) < 0) goto quit;   // major version
    if(fgetxx(fd, &n, 16, endian) < 0) goto quit;   // minor version
    if(fgetxx(fd, &n, 32, endian) < 0) goto quit;   // timezone offset
    if(fgetxx(fd, &n, 32, endian) < 0) goto quit;   // timestamp accuracy
    if(fgetxx(fd, &n, 32, endian) < 0) goto quit;   // snapshot length
    if(fgetxx(fd, &n, 32, endian) < 0) goto quit;   // link-layer type
    if_offset = pcap_hdrlen(n);

    hosts_len     = 0;
    hosts_memsize = 32;
    hosts = malloc(hosts_memsize * sizeof(hosts_t));
    if(!hosts) std_err();

    for(;;) {
        if(fgetxx(fd, &n, 32, endian) < 0) break;   // seconds
        if(fgetxx(fd, &n, 32, endian) < 0) break;   // microseconds
        if(fgetxx(fd, &n, 32, endian) < 0) break;   // caplen
        if(fgetxx(fd, &n, 32, endian) < 0) break;   // pcklen
        if(udp2tcp(fd, fdo, n) < 0) break;
    }

quit:
    fprintf(stderr, "- finished\n");
    if(fd != stdin)  fclose(fd);
    if(fd != stdout) fclose(fdo);
    if(hosts) free(hosts);
    return(0);
}



    // returns -1 only in case of file errors!!!
int udp2tcp(FILE *fd, FILE *fdo, u32 caplen) {
    iph         *ip;
    udph        *udp;
    tcph        *tcp;
    static u32  pktlen;
    u32         len,
                sip,
                dip;
    int         iflen,
                iplen;
    u16         sport,
                dport;
    static u8   *pkt   = NULL;
    u8          *data;

    if(!pkt) {
        pktlen = 256 + 0xffff;
        pkt    = malloc(pktlen);
        if(!pkt) std_err();
    }
    if(caplen > pktlen) {
        pktlen = caplen;
        pkt = realloc(pkt, pktlen);
        if(!pkt) std_err();
    }
    if(fread(pkt, caplen, 1, fd) != 1) return(-1);

    iflen  = ntohs(*(u16 *)(pkt + if_offset - 2));
    ip     = (iph *)(pkt + if_offset + ether_hdrlen(iflen));
    if((ip->ihl_ver >> 4) != 4) return(0);
    iplen  = (ip->ihl_ver & 15) << 2;

    sip    = ip->saddr;
    dip    = ip->daddr;

    if(ip->protocol == 17) {
        udp    = (udph *)((u8 *)ip + iplen);
        data   = (u8 *)((u8 *)udp + sizeof(udph));
        len    = ntohs(udp->len) - sizeof(udph);
        sport  = udp->source;
        dport  = udp->dest;

    } else if(ip->protocol == 6) {
        tcp    = (tcph *)((u8 *)ip + iplen);
        data   = (u8 *)((u8 *)tcp + (tcp->doff >> 2));
        len    = ntohs(ip->tot_len) - (data - (u8 *)ip);
        if(!(tcp->flags & TH_PSH)) return(0);
        sport  = tcp->source;
        dport  = tcp->dest;

    } else {
        return(0);
    }

    if(len <= 0) return(0);
    dump(fdo, sip, sport, dip, dport, data, len);

    return(0);
}



void dump(FILE *fdo, u32 sip, u16 sport, u32 dip, u16 dport, u8 *data, int datalen) {
    hosts_t *w;
    int     i;

    fprintf(stderr, "  %u bytes from %s:%hu to ",
        datalen,
        ip2str(sip), ntohs(sport));
    fprintf(stderr, "%s:%hu\n",
        ip2str(dip), ntohs(dport));

    for(i = 0; i < hosts_len; i++) {            // if already in the database
        w = &hosts[i];

        if((sip == w->sip) && (sport == w->sport) && (dip == w->dip) && (dport == w->dport)) {
            acp_dump(
                fdo, 6,
                sip, sport, dip, dport,
                data, datalen,
                &w->seq1, &w->ack1, &w->seq2, &w->ack2);
            return;
        }
        if((sip == w->dip) && (sport == w->dport) && (dip == w->sip) && (dport == w->sport)) {
            acp_dump(
                fdo, 6,
                sip, sport, dip, dport,
                data, datalen,
                &w->seq2, &w->ack2, &w->seq1, &w->ack1);
            return;
        }
    }

    if(hosts_len >= hosts_memsize) {            // if no space
        hosts_memsize += 32;
        hosts = realloc(hosts, hosts_memsize * sizeof(hosts_t));
        if(!hosts) std_err();
    }

    fprintf(stderr, "- new TCP connection\n");  // add new entry
    w = &hosts[i];

    w->sip      = sip;
    w->sport    = sport;
    w->dip      = dip;
    w->dport    = dport;
    w->seq1     = 0;
    w->ack1     = 0;
    w->seq2     = 0;
    w->ack2     = 0;
    hosts_len++;

    acp_dump_handshake(
        fdo, 6,
        sip, sport, dip, dport,
        NULL, 0,
        &w->seq1, &w->ack1, &w->seq2, &w->ack2);

    acp_dump(
        fdo, 6,
        sip, sport, dip, dport,
        data, datalen,
        &w->seq1, &w->ack1, &w->seq2, &w->ack2);
}



int fgetxx(FILE *fd, u32 *ret, int bits, int endian) {
    u32     num;
    int     i,
            bytes;
    u8      data[4];

    bytes = bits >> 3;

    if(fread(data, bytes, 1, fd) != 1) return(-1);

    for(num = i = 0; i < bytes; i++) {
        if(endian == LITTLE_ENDIAN) {
            num |= (data[i] << ((bytes - 1 - i) << 3));
        } else {
            num |= (data[i] << (i << 3));
        }
    }

    *ret = num;
    return(0);
}



void std_err(void) {
    perror("\nError");
    exit(1);
}


